define("epi-cms/project/ProjectComponentViewModel", [
// dojo
    "dojo/_base/declare",
    "dojo/_base/lang",
    "dojo/Deferred",
    "dojo/Evented",
    "dojo/Stateful",
    "dojo/topic",
    "dojo/dom-class",
    "dojo/promise/all",
    "dojo/when",

// dijit
    "dijit/Destroyable",

//epi
    "epi/epi",
    "epi/datetime",
    "epi/dependency",
    "epi/username",
    "epi-cms/core/ContentReference",
    "epi/shell/xhr/ErrorHandler",
    "epi-cms/contentediting/ContentActionSupport",

// Commands
    "epi/shell/command/withConfirmation",
    "./command/AddProject",
    "./command/PublishProject",
    "./command/ReadyToPublishProjectItem",
    "./command/RefreshProjectItems",
    "./command/RemoveProject",
    "./command/RemoveProjectItem",
    "./command/RenameProject",
    "./command/SortProjectItems",

// Resources
    "epi/i18n!epi/nls/episerver.cms.components.project",
    "epi/i18n!epi/nls/episerver.shared.action",
    "epi/i18n!epi/cms/nls/episerver.cms.contentediting.versionstatus"
],

function (
// dojo
    declare,
    lang,
    Deferred,
    Evented,
    Stateful,
    topic,
    domClass,
    all,
    when,

// dijit
    Destroyable,

// epi
    epi,
    epiDate,
    dependency,
    username,
    ContentReference,
    ErrorHandler,
    ContentActionSupport,

// Commands
    withConfirmation,
    AddProject,
    PublishProject,
    ReadyToPublishProjectItem,
    RefreshProjectItems,
    RemoveProject,
    RemoveProjectItem,
    RenameProject,
    SortProjectItems,

// Resources
    res,
    resShared,
    versionStatusRes
) {
    return declare([Stateful, Destroyable, Evented], {
        // summary:
        //      The view model for the epi-cms/project/ProjectComponent
        // tags:
        //      internal

        // commands: [readonly] epi/shell/command/_Command[]
        //      Commands to be consumed by the view.
        commands: null,

        // namedCommands: [readonly] Object
        //      Way to access commands using named keys.
        namedCommands: null,

        canBePublished: true,

        // created: [readonly] string
        //      A view prepared version of the date the selected project was created.
        created: "",

        // createdBy: [readonly] string
        //      A view prepared version of the user that created the selected project.
        createdBy: "",

        // profile: [readonly] Profile
        //      The current user profile
        profile: null,

        // createdProjectInfo: [readonly] string
        //      A view prepared version of the placeholder info.
        createProjectInfo: "",

        // totalProjectItems: [readonly] int
        //      Current project items count.
        totalProjectItems: null,
        
        // hasProject: [public] bool
        //      True if we have any projects
        hasProject: null,

        // projectStore: [readonly] Store
        //      A REST store for interacting with projects.
        projectStore: null,

        // projectItemStore: [readonly] Store
        //      A REST store for interacting with project items.
        projectItemStore: null,

        // projectItemQuery: Query
        //      Query object holding parameters to get a projects items.
        projectItemQuery: null,

        // projectItemSortOrder: [public] String
        //      The order to sort projects items by.
        projectItemSortOrder: null,

        // selectedProject: [public] Object
        //      The project that is currently selected.
        selectedProject: null,

        // selectedProjectItem: [public] Object
        //      The project item that is currently selected.
        selectedProjectItem: null,

        _isPublishingHandle : null,

        _selectedProjectProfileKey: "epi.selected-project-id",

        _sortOrderProfileKey: "epi.project-sort-order",

        _selectedProjectSetter: function(selectedProject) {
            var created = "",
                createdBy = "";

            // Early exit if the project has not changed
            if (epi.areEqual(this.selectedProject, selectedProject)) { return; }

            this.selectedProject = selectedProject;
            // Refresh the list query when the selected project changes.
            this.set("projectItemQuery", selectedProject ? { projectId: selectedProject.id } : null);

            this.set("canBePublished", selectedProject && selectedProject.canBePublished);

            if (selectedProject) {
                created = epiDate.toUserFriendlyString(epiDate.deserialize(selectedProject.created));
                createdBy = username.toUserFriendlyString(selectedProject.createdBy, null, true);
            }

            this.set("created", created);
            this.set("createdBy", createdBy);

            //Store the selected project
            this._storeSelectedProjectInProfile();
        },

        postscript: function () {
            this.inherited(arguments);

            // Resolve the stores from the registry if they haven't been injected. This allows
            // us to mock the stores when testing.
            this.projectStore = this.projectStore || dependency.resolve("epi.storeregistry").get("epi.cms.project");
            this.projectItemStore = this.projectItemStore || dependency.resolve("epi.storeregistry").get("epi.cms.project.item");
            this.profile = this.profile || dependency.resolve("epi.shell.Profile");

            this._createCommands();

            // Set the sort order based on what is stored in the profile, fall back to
            // default value if nothing stored in profile.
            this.loadSortOrderFromProfile()
                // Ensure that we set sort before the selected project so that we don't end up with two requests to the server.
                .then(lang.hitch(this, function(sortOrder) {
                    this._changeAttrValue("projectItemSortOrder", sortOrder || "+Name");
                }))
                // Load the last selected project from the profile regardless of whether the sort order was set.
                .always(lang.hitch(this, "loadSelectedProjectFromProfile"));
        },

        loadSelectedProjectFromProfile: function() {
            // summary:
            //      Load the selected project from the user profile
            // returns: Promise
            //      The project that was stored in the profile
            // tags:
            //      internal

            var self = this,
                def = this._loadSelectedProjectProfilePromise = new Deferred();

            //Load the selected project from the profile
            when(this.profile.get(this._selectedProjectProfileKey), function(projectId) {
                if(!projectId) {
                    def.resolve();
                    return;
                }

                // Need to get the project from the store in order to ensure it still exists.
                when(self.projectStore.get(projectId), function(project) {
                    //Set the selected project
                    if(project) {
                        self.set("selectedProject", project);
                    }
                    def.resolve(project);
                }, def.reject);
            }, def.reject);

            return def.promise;
        },

        loadSortOrderFromProfile: function() {
            // summary:
            //      Load the sort order from the user profile.
            // returns: Promise
            //      A promise that resolves with the sort order stored in the profile or null.
            // tags:
            //      internal
            var deferred = new Deferred();

            when(this.profile.get(this._sortOrderProfileKey), deferred.resolve, deferred.reject);

            return deferred.promise;
        },

        getCommands: function () {
            // summary:
            //      Returns all available commands
            // tags:
            //      public

            return this.commands;
        },

        addProject: function (project) {
            // summary:
            //      Adds a project and sets it as the selectedProject
            // tags:
            //      public

            this.projectStore.add(project)
                .then(lang.hitch(this, function(value) { this.set("selectedProject", value); }))
                .otherwise(ErrorHandler.forXhr);
        },

        updateProject: function(project) {
            // summary:
            //      Adds a project and sets it as the selectedProject
            // tags:
            //      public

            this.projectStore.put(project)
                .then(lang.hitch(this, "set", "selectedProject"))
                .otherwise(ErrorHandler.forXhr);
        },

        selectNextProject: function (sort) {
            // ummary:
            //      Selects next project
            // tags:
            //      public

            this._getNextProject(0,1,sort).then(lang.hitch(this, function (results) {
                var value = results[0];
                this.set("selectedProject", value);
            }));
        },

        publishProject: function (selectedProjectId) {
            // summary:
            //      Starts a publishing task for a project and polls continously to verify the results
            // tags:
            //      public

            var def = new Deferred(),
                store = this.projectStore,
                self = this;

            if (isNaN(selectedProjectId)) {
                throw new Error("selectedProjectId is not a number");
            }

            store.executeMethod("Publish", selectedProjectId).then(
                function (updatedProject) {
                    // Call notify on the store once the publish method is complete to
                    // ensure any observes are kept up to date.
                    store.notify(updatedProject, store.getIdentity(updatedProject));

                    self.set("selectedProject", updatedProject);
                    def.resolve(updatedProject);
                },
                function() {
                    def.reject("publishFailed");
                }
            );

            return def.promise;
        },

        refreshProject: function() {
            // summary:
            //      Refreshes the selected project, particularly information if the project can be published or not.
            // tags:
            //      public

            this._updateHasProject();

            var self = this;

            if(!this.selectedProject) {
                this.emit("refresh");
                return;
            }

            return this.projectStore.refresh(this.selectedProject.id).then(
                function(value) {
                    self.set("selectedProject", value);
                    self.emit("refresh");
                },
                function (ex) {
                    self.set("selectedProject", null);
                    console.error("Failed to update selectedProject", ex);
                }
            );
        },

        addProjectItems: function(contentReferences, projectId) {
            // summary:
            //      Adds the given content references as items of the project.
            // tags:
            //      public

            // Ensure the content reference we received are valid.
            contentReferences = contentReferences.filter(function(reference) {
                if (ContentReference.isContentReference(reference)) {
                    return true;
                }
                console.log("The given value '" + reference + "' is not a valid content reference.");
            });

            if (!projectId && !this.selectedProject) {
                throw new Error("If no projectId is provided a selectedProject must be set in the context");
            }

            projectId = projectId || this.get("selectedProject").id;

            var store = this.projectItemStore,
                promises = contentReferences.map(function(reference) {
                    return store.add({ contentLink: reference, projectId: projectId });
                });

            return all(promises)
                .then(lang.hitch(this, "emit", "refresh"))
                .otherwise(ErrorHandler.forXhr);
        },

        removeSelectedProjectItem: function() {
            // summary:
            //      Removes a project item from a project.
            // tags:
            //      public

            this.projectItemStore.remove(this.selectedProjectItem.id)
                .then(lang.hitch(this, function() {
                    this.set("totalProjectItems", (this.totalProjectItems - 1));
                }))
                .otherwise(ErrorHandler.forXhr);
        },

        markProjectItemAsReadyToPublish: function() {
            // summary:
            //      Mark the project item as ready to publish
            // returns: Promise
            //      A promise that resolves when the refresh project has completed.
            // tags:
            //      public

            var promise = this.projectItemStore.executeMethod("ReadyToPublish", this.selectedProjectItem.id);

            promise
                .then(lang.hitch(this, "refreshProject"))
                .otherwise(ErrorHandler.forXhr);

            return promise;
        },

        requestContextChange: function (contentReference) {
            // summary:
            //      Requests that the UI changes the context to the specified contentReference.
            // tags:
            //      public

            topic.publish("/epi/shell/context/request", { uri: "epi.cms.contentdata:///" + contentReference }, { sender: this });
        },

        _createCommands: function () {

            var commandArgs = { model: this },
                refreshProjectItems,
                addProject,
                publishProject,
                readyToPublishProjectItem,
                removeProject,
                removeProjectItem,
                renameProject,
                sortProjectItems = new SortProjectItems(commandArgs);

            this.own(
                refreshProjectItems = new RefreshProjectItems(commandArgs),
                addProject = new AddProject(commandArgs),
                publishProject = new PublishProject(commandArgs),
                readyToPublishProjectItem = new ReadyToPublishProjectItem(commandArgs),
                removeProject = withConfirmation(new RemoveProject({ model: this, store: this.projectStore }), null, {
                    title: res.command.removeproject.label,
                    description: res.command.removeproject.confirmation,
                    confirmActionText: resShared.deletelabel,
                    cancelActionText: resShared.cancel
                }),
                removeProjectItem = withConfirmation(new RemoveProjectItem({ model: this, store: this.projectItemStore }), null, {
                    title: res.command.removeprojectitem.label,
                    description: res.command.removeprojectitem.confirmation,
                    confirmActionText: resShared.remove,
                    cancelActionText: resShared.cancel
                }),
                renameProject = new RenameProject(commandArgs)
            );

            this.set("commands", [
                refreshProjectItems,
                addProject,
                renameProject,
                publishProject,
                removeProject,
                readyToPublishProjectItem,
                removeProjectItem,
                sortProjectItems
            ]);

            this.set("namedCommands", {
                addProject: addProject,
                refreshProjectItems: refreshProjectItems,
                publishProject: publishProject,
                removeProject: removeProject,
                removeProjectItem: removeProjectItem
            });
        },

        statusFormatter: function(value, object, node, options) {
            // summary:
            //      Usable for the view to get a icon indicating the status of a item.
            // tags:
            //      public

            var accessLevel = ContentActionSupport.accessLevel,
                versionStatus = [
                    "",                                 // NotCreated
                    "Stop epi-icon--danger",            // Rejected
                    "Pen",                              // CheckedOut
                    "Checkmark epi-icon--success",      // CheckedIn
                    "Published epi-icon--success",      // Published
                    "PreviouslyPublished",              // PreviouslyPublished
                    "Clock",                            // DelayedPublish
                    "Clock epi-icon--danger"            // Expired
                ],
                cssClass = "epi-icon" + versionStatus[object.status],
                title = ContentActionSupport.versionLocalizations[object.status] || "";

            // of the content item is deleted set the deleted class on the entire node
            if (object.isDeleted) {
                cssClass = "epi-iconWarning epi-icon--warning";
                title = versionStatusRes.versionnotfound;
                domClass.add(node, "epi-content--deleted");
            } else if (!ContentActionSupport.hasAccess(object.accessMask, accessLevel.Read)) {
                cssClass = "epi-iconLock";
                title = res.status.message.noreadaccess;
            } else if (!ContentActionSupport.hasAccess(object.accessMask, accessLevel.Publish)) {
                cssClass = "epi-iconWarning epi-icon--warning";
                title = res.status.message.nopublishaccess;
            }

            return '<span class="dijitReset dijitInline dijitIcon epi-objectIcon ' + cssClass + '" title="' + title + '"></span>' + value;
        },

        _getNextProject: function (start, count, sort) {
            // summary:
            //      Gets projects based on attributes
            // tags:
            //      private
            return this.projectStore.query(null, {
                start: start,
                count: count,
                sort: sort
            });
        },

        _updateHasProject: function () {
            // summary:
            //      updates hasProject boolean depending on getHasProject result
            // tags:
            //      private
            this._getNextProject(0,1).then(
                lang.hitch(this, function (results) {
                    var result = results.length > 0;
                    this.set("hasProject", result);
                })
            );
        },

        _storeSelectedProjectInProfile: function() {
            // summary:
            //      Store the selected project in the user profile

            //If we haven't loaded the selected profile yet (or is ongoing) do not try to save it. E.g. at startup
            if(this._loadSelectedProjectProfilePromise && !this._loadSelectedProjectProfilePromise.isFulfilled()) {
                return;
            }
            var project = this.get("selectedProject");

            return when(this.profile.set(this._selectedProjectProfileKey, project && project.id, {
                location: "server"
            }));
        },

        _projectItemSortOrderSetter: function(sortOrder) {
            // summary:
            //      Store the selected project in the user profile
            this.projectItemSortOrder = sortOrder;

            return this.profile.set(this._sortOrderProfileKey, sortOrder, { location: "server" });
        }
    });
});
